use crate::{
    get_user_from_token,
    model::{ApiReturn, Error},
    routes::{api::v1::CreateUserWarning, pages::PaginatedQuery},
    State,
};
use axum::{
    extract::{Path, Query},
    response::IntoResponse,
    Extension, Json,
};
use axum_extra::extract::CookieJar;
use tetratto_core::model::{auth::UserWarning, oauth, permissions::FinePermission};

/// Create a new user warning.
pub async fn create_request(
    jar: CookieJar,
    Path(uid): Path<usize>,
    Extension(data): Extension<State>,
    Json(req): Json<CreateUserWarning>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::ModManageWarnings) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    if !user.permissions.check(FinePermission::MANAGE_BANS) {
        return Json(Error::NotAllowed.into());
    }

    match data
        .create_user_warning(UserWarning::new(uid, user.id, req.content))
        .await
    {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "User warning created".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

/// Delete the given user warning.
pub async fn delete_request(
    jar: CookieJar,
    Path(id): Path<usize>,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::ModManageWarnings) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    if !user.permissions.check(FinePermission::MANAGE_WARNINGS) {
        return Json(Error::NotAllowed.into());
    }

    match data.delete_user_warning(id, user).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "User warning deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

/// Get all warnings for the given user.
pub async fn on_user_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Query(props): Query<PaginatedQuery>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::ModManageWarnings) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    if !user.permissions.check(FinePermission::MANAGE_WARNINGS) {
        return Json(Error::NotAllowed.into());
    }

    match data
        .get_user_warnings_by_user(user.id, 12, props.page)
        .await
    {
        Ok(w) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(w),
        }),
        Err(e) => Json(e.into()),
    }
}

/// Get a single warning.
pub async fn get_request(
    jar: CookieJar,
    Path(id): Path<usize>,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::ModManageWarnings) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    if !user.permissions.check(FinePermission::MANAGE_WARNINGS) {
        return Json(Error::NotAllowed.into());
    }

    match data.get_user_warning_by_id(id).await {
        Ok(w) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(w),
        }),
        Err(e) => Json(e.into()),
    }
}
